package ch.psi.daq.queryrest.controller;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.http.MediaType;
import org.springframework.validation.Validator;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.InitBinder;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;

import ch.psi.daq.cassandra.config.CassandraConfig;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.json.status.channel.ChannelStatus;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.ChannelsResponse;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.query.operation.Response;
import ch.psi.daq.domain.query.operation.ResponseFormat;
import ch.psi.daq.domain.query.status.channel.ChannelStatusQuery;
import ch.psi.daq.domain.reader.Backend;
import ch.psi.daq.domain.request.validate.RequestProviderValidator;
import ch.psi.daq.domain.status.StatusReader;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.processor.ChannelNameCache;
import ch.psi.daq.query.processor.QueryProcessor;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.query.QueryManager;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;
import ch.psi.daq.queryrest.response.PolymorphicResponseMixIn;
import ch.psi.daq.queryrest.response.json.JSONHTTPResponse;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.Lists;

@RestController
public class QueryRestController {

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestController.class);

   public static final String PATH_CHANNELS = DomainConfig.PATH_CHANNELS;
   public static final String PATH_QUERY = DomainConfig.PATH_QUERY;
   public static final String PATH_QUERIES = DomainConfig.PATH_QUERIES;
   public static final String PATH_STATUS_CHANNELS = DomainConfig.PATH_STATUS_CHANNELS;

   @Resource
   private ApplicationContext appContext;

   @Resource
   private Validator queryValidator;
   private Validator requestProviderValidator = new RequestProviderValidator();

   @Resource
   private QueryManager queryManager;

   @Resource
   private ObjectMapper objectMapper;

   private Response defaultResponse = new JSONHTTPResponse();

   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;

   @Resource(name = QueryRestConfig.BEAN_NAME_READ_TIMOUT)
   private Integer readTimeout;

   private Map<Backend, QueryProcessor> queryProcessors = new LinkedHashMap<>();
   private ChannelNameCache channelNameCache;

   private Map<Backend, StatusReader> statusReaders = new LinkedHashMap<>();

   @PostConstruct
   public void afterPropertiesSet() {
      List<Exception> exceptions = new ArrayList<>();

      try {
         QueryProcessor queryProcessor =
               appContext.getBean(QueryConfig.BEAN_NAME_CASSANDRA_QUERY_PROCESSOR, QueryProcessor.class);
         queryProcessors.put(queryProcessor.getBackend(), queryProcessor);
      } catch (Exception e) {
         exceptions.add(e);
         LOGGER.warn("");
         LOGGER.warn("##########");
         LOGGER.warn("Could not load query processor for cassandra.");
         LOGGER.warn("##########");
         LOGGER.warn("");
      }

      try {
         QueryProcessor queryProcessor =
               appContext.getBean(QueryConfig.BEAN_NAME_ARCHIVER_APPLIANCE_QUERY_PROCESSOR, QueryProcessor.class);
         queryProcessors.put(queryProcessor.getBackend(), queryProcessor);
      } catch (Exception e) {
         exceptions.add(e);
         LOGGER.warn("");
         LOGGER.warn("##########");
         LOGGER.warn("Could not load query processor for archiverappliance.");
         LOGGER.warn("##########");
         LOGGER.warn("");
      }

      if (queryProcessors.isEmpty()) {
         LOGGER.error("No query processor could be loaded! Exceptions were: ");
         for (Exception exception : exceptions) {
            LOGGER.error("", exception);
         }

         throw new RuntimeException("No Backends available!");
      }

      try {
         StatusReader statusReader =
               appContext.getBean(QueryConfig.BEAN_NAME_CASSANDRA_STATUS_READER, StatusReader.class);
         statusReaders.put(statusReader.getBackend(), statusReader);
      } catch (Exception e) {
         exceptions.add(e);
         LOGGER.warn("");
         LOGGER.warn("##########");
         LOGGER.warn("Could not load status reader for cassandra.");
         LOGGER.warn("##########");
         LOGGER.warn("");
      }

      try {
         StatusReader statusReader =
               appContext.getBean(QueryConfig.BEAN_NAME_ARCHIVER_APPLIANCE_STATUS_READER, StatusReader.class);
         statusReaders.put(statusReader.getBackend(), statusReader);
      } catch (Exception e) {
         exceptions.add(e);
         LOGGER.warn("");
         LOGGER.warn("##########");
         LOGGER.warn("Could not load status reader for archiverappliance.");
         LOGGER.warn("##########");
         LOGGER.warn("");
      }

      channelNameCache =
            new ChannelNameCache(queryProcessors, appContext.getBean(CassandraConfig.BEAN_NAME_READ_TIMEOUT,
                  Integer.class).longValue());
   }

   @PreDestroy
   public void destroy() {
      channelNameCache.destroy();
   }

   @InitBinder
   protected void initBinder(WebDataBinder binder) {
      if (binder.getTarget() != null) {
         if (requestProviderValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(requestProviderValidator);
         }
         if (queryValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(queryValidator);
         }
      }
   }

   @RequestMapping(value = PATH_CHANNELS, method = {RequestMethod.GET, RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ChannelsResponse> getChannels(@RequestBody(required = false) ChannelsRequest request)
         throws Throwable {
      return queryManager.getChannels(request);
   }

   /**
    * Query specific channel names, and return only those.
    * 
    * @param channelName part of (or full) channel name
    * @return Collection of channel names matching the specified input channel name
    * @throws Throwable in case something goes wrong
    */
   @RequestMapping(value = PATH_CHANNELS + "/{channelName}", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<ChannelsResponse> getChannels(@PathVariable(value = "channelName") String channelName)
         throws Throwable {
      return getChannels(new ChannelsRequest(channelName));
   }

   /**
    * Queries for channels status
    * 
    * @param query the {@link ChannelStatusQuery}
    * @param res the {@link HttpServletResponse} instance associated with this request
    * @return Collection of channel status info
    * @throws Throwable in case something goes wrong
    */
   @RequestMapping(
         value = PATH_STATUS_CHANNELS,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<ChannelStatus> executeChannelStatusQuery(@RequestBody ChannelStatusQuery query,
         HttpServletResponse res) throws Throwable {

      // set backends if not defined yet
      channelNameCache.setBackends(query.getChannels());

      List<CompletableFuture<ChannelStatus>> futures = new ArrayList<>(query.getChannels().size());
      for (ChannelName channel : query.getChannels()) {
         StatusReader statusReader = statusReaders.get(channel.getBackend());
         if (statusReader != null) {
            futures.add(statusReader.getChannelStatusAsync(channel.getName()));
         } else {
            LOGGER.warn("There is no StatusReader available for '{}'.", channel.getBackend());
         }
      }

      try {
         List<ChannelStatus> retStatus = new ArrayList<>(futures.size());
         for (CompletableFuture<ChannelStatus> completableFuture : futures) {
            retStatus.add(completableFuture.get(readTimeout, TimeUnit.SECONDS));
         }
         return retStatus;
      } catch (Exception e) {
         String message = "Could not extract ChannelStatus within timelimits.";
         LOGGER.error(message, e);
         throw new IllegalStateException(message, e);
      }
   }

   /**
    * Accepts the properties for the {@link DAQQuery} instance as stringified JSON query string.
    * 
    * @param jsonBody The {@link DAQQuery} properties sent as a JSON string, i.e. this is the
    *        stringified body of the POST request method
    * @param res the current {@link HttpServletResponse} instance
    * @throws Exception if reading the JSON string fails or if the subsequent call to
    *         {@link #executeQuery(DAQQuery, HttpServletResponse)} fails
    */
   @RequestMapping(
         value = PATH_QUERY,
         method = RequestMethod.GET)
   public void executeQueryBodyAsString(@RequestParam String jsonBody, HttpServletResponse res) throws Exception {
      DAQQuery query = objectMapper.readValue(jsonBody, DAQQuery.class);
      executeQuery(query, res);
   }

   /**
    * Executes a single query.
    * 
    * @param query the {@link DAQQuery}
    * @param res the {@link HttpServletResponse} instance associated with this request
    * @throws Exception thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = PATH_QUERY,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQuery(@RequestBody @Valid DAQQuery query, HttpServletResponse res) throws Exception {
      executeQueries(new DAQQueries(query), res);
   }

   /**
    * Accepts the properties for the {@link DAQQueries} instance as stringified JSON query string.
    * 
    * @param jsonBody The {@link DAQQueries} properties sent as a JSON string, i.e. this is the
    *        stringified body of the POST request method
    * @param res the current {@link HttpServletResponse} instance
    * @throws Exception if reading the JSON string fails or if the subsequent call to
    *         {@link #executeQueries(DAQQueries, HttpServletResponse)} fails
    */
   @RequestMapping(
         value = PATH_QUERIES,
         method = RequestMethod.GET)
   public void executeQueriesBodyAsString(@RequestParam String jsonBody, HttpServletResponse res) throws Exception {
      DAQQueries queries = objectMapper.readValue(jsonBody, DAQQueries.class);
      executeQueries(queries, res);
   }

   /**
    * Catch-all query method for getting data from the backend for both JSON and CSV requests.
    * <p>
    * The {@link DAQQueries} object will contain the concrete subclass based on the combination of
    * fields defined in the user's query. The AttributeBasedDeserializer decides which class
    * to deserialize the information into and has been configured (see
    * QueryRestConfig#afterPropertiesSet) accordingly.
    * 
    * @param queries the {@link DAQQueries}
    * @param res the {@link HttpServletResponse} instance associated with this request
    * @throws Exception thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = PATH_QUERIES,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQueries(@RequestBody @Valid DAQQueries queries, HttpServletResponse res) throws Exception {
      try {
         LOGGER.debug("Executing queries '{}'", queries);

         Response response = queries.getResponseOrDefault(defaultResponse);
         if (response instanceof AbstractHTTPResponse) {
            ((AbstractHTTPResponse) response).respond(appContext, queries, res);
         } else {
            String message =
                  String.format(
                        "Expecting Response of type '%s' but received '%s'. Check JSON deserialization defined in '%s'",
                        AbstractHTTPResponse.class.getName(), response.getClass().getName(),
                        PolymorphicResponseMixIn.class.getName());
            LOGGER.error(message);
            throw new IllegalArgumentException(message);
         }


      } catch (Exception e) {
         LOGGER.error("Failed to execute query '{}'.", queries, e);
         throw e;
      }
   }

   /**
    * Returns the current list of {@link Ordering}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @RequestMapping(value = "ordering", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Ordering> getOrderingValues() {
      return Lists.newArrayList(Ordering.values());
   }

   /**
    * Returns the current list of {@link ResponseFormat}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @RequestMapping(value = "responseformat", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ResponseFormat> getResponseFormatValues() {
      return Lists.newArrayList(ResponseFormat.values());
   }

   /**
    * Returns the current list of {@link QueryField}s available.
    * 
    * @return list of {@link QueryField}s as String array
    */
   @RequestMapping(value = "queryfields", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<QueryField> getQueryFieldValues() {
      return Lists.newArrayList(QueryField.values());
   }

   /**
    * Returns the current list of {@link Aggregation}s available.
    * 
    * @return list of {@link Aggregation}s as String array
    */
   @RequestMapping(value = "aggregations", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Aggregation> getAggregationsValues() {
      return Lists.newArrayList(Aggregation.values());
   }

   /**
    * Returns the current list of {@link AggregationType}s available.
    * 
    * @return list of {@link AggregationType}s as String array
    */
   @RequestMapping(value = "aggregationtypes", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<AggregationType> getAggregationTypesValues() {
      return Lists.newArrayList(AggregationType.values());
   }

   /**
    * Returns the current list of {@link Backend} values.
    * 
    * @return list of {@link Backend}s as String array
    */
   @RequestMapping(value = "backends", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Backend> getDBModeValues() {
      return Lists.newArrayList(Backend.values());
   }
}
