package ch.psi.daq.queryrest.controller.validator;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.Set;

import javax.annotation.Resource;

import org.springframework.validation.Errors;
import org.springframework.validation.Validator;

import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.request.Request;
import ch.psi.daq.queryrest.config.QueryRestConfig;

public class QueryValidator implements Validator {

   @Resource(name = QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_FIELDS)
   private Set<QueryField> defaultResponseFields;

   @Resource(name = QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS)
   private Set<Aggregation> defaultResponseAggregations;

   /**
    * {@inheritDoc}
    */
   @Override
   public boolean supports(Class<?> clazz) {
      return DAQQuery.class.isAssignableFrom(clazz) || DAQQueries.class.isAssignableFrom(clazz);
   }

   /**
    * {@inheritDoc}
    */
   @Override
   public void validate(Object target, Errors errors) {
      if (target instanceof DAQQuery) {
         this.checkElement((DAQQuery) target, errors);
      } else if (target instanceof DAQQueries) {
         DAQQueries queries = (DAQQueries) target;
         for (DAQQueryElement daqQueryElement : queries) {
            this.checkElement(daqQueryElement, errors);
         }
      }
   }

   private void checkElement(DAQQueryElement query, Errors errors) {
      // set default values (if not set)
      if (query.getFields() == null || query.getFields().isEmpty()) {
         query.setFields(new LinkedHashSet<>(defaultResponseFields));
      }

      if (query.getAggregation() != null) {
         // check if only one binning element is defined
         long durationPerBin = query.getAggregation().getDurationPerBin();
         long pulsesPerBin = query.getAggregation().getPulsesPerBin();
         int nrOfBins = query.getAggregation().getNrOfBins();
         if ((durationPerBin != Request.NOT_SET && (pulsesPerBin != Request.NOT_SET || nrOfBins != Request.NOT_SET))
               || (pulsesPerBin != Request.NOT_SET && (durationPerBin != Request.NOT_SET || nrOfBins != Request.NOT_SET))
               || (nrOfBins != Request.NOT_SET && (durationPerBin != Request.NOT_SET || pulsesPerBin != Request.NOT_SET))) {
            errors.reject("durationPerBin", "Only one binning element must be defined.");
            errors.reject("pulsesPerBin", "Only one binning element must be defined.");
            errors.reject("nrOfBins", "Only one binning element must be defined.");
         }

         if (query.getRange().isPulseIdRangeDefined() && durationPerBin != Request.NOT_SET) {
            errors.reject("durationPerBin", "Pulse range queries only support pulse based binning.");
         }
         if (query.getRange().isTimeRangeDefined() && pulsesPerBin != Request.NOT_SET) {
            errors.reject("pulsesPerBin", "Time range queries only support time based binning.");
         }


         // set default values (if not set)
         if (query.getAggregation().getAggregations() == null || query.getAggregation().getAggregations().isEmpty()) {
            query.getAggregation().setAggregations(new ArrayList<>(defaultResponseAggregations));
         }
      }
   }
}
