package ch.psi.daq.queryrest.query;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;

import ch.psi.daq.cassandra.config.CassandraConfig;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.ChannelsResponse;
import ch.psi.daq.domain.reader.Backend;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.impl.BackendQuery;
import ch.psi.daq.query.processor.ChannelNameCache;
import ch.psi.daq.query.processor.QueryProcessor;

public class QueryManagerImpl implements QueryManager {
   private static final Logger LOGGER = LoggerFactory.getLogger(QueryManagerImpl.class);
   
   @Resource
   private ApplicationContext appContext;
   
   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;
   
   private Map<Backend, QueryProcessor> queryProcessors = new LinkedHashMap<>();
   private ChannelNameCache channelNameCache;

   @PostConstruct
   public void afterPropertiesSet() {
      List<Exception> exceptions = new ArrayList<>();

      try {
         QueryProcessor queryProcessor =
               appContext.getBean(QueryConfig.BEAN_NAME_CASSANDRA_QUERY_PROCESSOR, QueryProcessor.class);
         queryProcessors.put(queryProcessor.getBackend(), queryProcessor);
      } catch (Exception e) {
         exceptions.add(e);
         LOGGER.warn("");
         LOGGER.warn("##########");
         LOGGER.warn("Could not load query processor for cassandra.");
         LOGGER.warn("##########");
         LOGGER.warn("");
      }

      try {
         QueryProcessor queryProcessor =
               appContext.getBean(QueryConfig.BEAN_NAME_ARCHIVER_APPLIANCE_QUERY_PROCESSOR, QueryProcessor.class);
         queryProcessors.put(queryProcessor.getBackend(), queryProcessor);
      } catch (Exception e) {
         exceptions.add(e);
         LOGGER.warn("");
         LOGGER.warn("##########");
         LOGGER.warn("Could not load query processor for archiverappliance.");
         LOGGER.warn("##########");
         LOGGER.warn("");
      }

      if (queryProcessors.isEmpty()) {
         LOGGER.error("No query processor could be loaded! Exceptions were: ");
         for (Exception exception : exceptions) {
            LOGGER.error("", exception);
         }

         throw new RuntimeException("No Backends available!");
      }

      channelNameCache =
            new ChannelNameCache(queryProcessors, appContext.getBean(CassandraConfig.BEAN_NAME_READ_TIMEOUT,
                  Integer.class).longValue());
   }
   
   @PreDestroy
   public void destroy() {
      channelNameCache.destroy();
   }
   
   @Override
   public List<ChannelsResponse> getChannels(ChannelsRequest request) {
      // in case not specified use defaults (e.g. GET)
      if (request == null) {
         request = new ChannelsRequest();
      }

      return channelNameCache.getChannels(request);
   }
   
   @Override
   public List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> executeQueries(DAQQueries queries) {
      // set backends if not defined yet
      channelNameCache.setBackends(queries);

      List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results =
            new ArrayList<>(queries.getQueries().size());

      for (DAQQueryElement queryElement : queries) {
         Stream<Triple<BackendQuery, ChannelName, ?>> resultStreams =
               BackendQuery
                     .getBackendQueries(queryElement)
                     .stream()
                     .filter(query -> {
                        QueryProcessor processor = queryProcessors.get(query.getBackend());
                        if (processor != null) {
                           return true;
                        } else {
                           LOGGER.warn("There is no QueryProcessor available for '{}'", query.getBackend());
                           return false;
                        }
                     })
                     .flatMap(query -> {
                        QueryProcessor processor = queryProcessors.get(query.getBackend());
                        QueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);

                        // all the magic happens here
                           Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEvents =
                                 processor.process(queryAnalizer);
                           // do post-process
                           Stream<Entry<ChannelName, ?>> channelToData = queryAnalizer.postProcess(channelToDataEvents);

                           return channelToData.map(entry -> {
                              return Triple.of(query, entry.getKey(), entry.getValue());
                           });
                        });

         // Now we have a stream that loads elements sequential BackendQuery by BackendQuery.
         // By materializing the outer Stream the elements of all BackendQuery are loaded async
         // (speeds things up but requires also more memory - i.e. it relies on Backends not loading
         // all elements into memory at once)
         resultStreams = resultStreams.collect(Collectors.toList()).stream();

         results.add(Pair.of(queryElement, resultStreams));
      }

      return results;
   }

}
