package ch.psi.daq.queryrest.response.csv;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.function.ToLongFunction;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.ServletResponse;

import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVPrinter;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.psi.daq.common.stream.StreamIterable;
import ch.psi.daq.common.stream.StreamMatcher;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.impl.BackendQuery;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;

/**
 * Takes a Java 8 stream and writes it to the output stream provided by the {@link ServletResponse}
 * of the current request.
 */
public class CSVResponseStreamWriter implements ResponseStreamWriter {
   private static final Logger LOGGER = LoggerFactory.getLogger(CSVResponseStreamWriter.class);

   public static final char DELIMITER_CVS = ';';
   public static final String DELIMITER_ARRAY = ",";
   public static final char DELIMITER_CHANNELNAME_FIELDNAME = '.';
   public static final String EMPTY_VALUE = "";
   public static final String FIELDNAME_EXTREMA = "extrema";
   private static final Function<Pair<ChannelName, DataEvent>, ChannelName> KEY_PROVIDER = (pair) -> pair.getKey();
   // try to match sync data (bsread) with non sync data (epics) based on the time usin 10 millis
   // buckets.
   private static final ToLongFunction<Pair<ChannelName, DataEvent>> MATCHER_PROVIDER = (pair) -> pair.getValue()
         .getGlobalMillis() / 10L;

   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;

   @Override
   public void respond(final List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results,
         final OutputStream out) throws Exception {
      AtomicReference<Exception> exception = new AtomicReference<>();

      final Map<ChannelName, Stream<Pair<ChannelName, DataEvent>>> streams = new LinkedHashMap<>(results.size());
      final List<String> header = new ArrayList<>();
      final Collection<Pair<ChannelName, Function<DataEvent, String>>> accessors = new ArrayList<>();

      // prepare the streams
      results.forEach(entry -> {
         final DAQQueryElement query = entry.getKey();

         entry.getValue()
               .sequential()
               .forEach(triple -> {
                  if (triple.getRight() instanceof Stream) {
                     setupChannelColumns(query, triple.getLeft(), triple.getMiddle(), header, accessors);

                     final Stream<Pair<ChannelName, DataEvent>> eventStream = ((Stream<DataEvent>) triple.getRight())
                           .map(dataEvent -> Pair.of(triple.getMiddle(), dataEvent));
                     streams.put(triple.getMiddle(), eventStream);
                  } else {
                     final String message = String.format("Expect a DataEvent Stream for '%s'.", triple.getMiddle());
                     LOGGER.warn(message);
                  }
               });
      });

      // online matching of the stream's content
      StreamMatcher<ChannelName, Pair<ChannelName, DataEvent>> streamMatcher =
            new StreamMatcher<>(KEY_PROVIDER, MATCHER_PROVIDER, streams.values());
      Iterator<Map<ChannelName, Pair<ChannelName, DataEvent>>> streamsMatchIter = streamMatcher.iterator();

      // prepare csv output
      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(DELIMITER_CVS);
      Writer writer = null;
      CSVPrinter csvFilePrinter = null;

      try {
         writer = new BufferedWriter(new OutputStreamWriter(out, StandardCharsets.UTF_8));
         // writer = new BufferedWriter(new OutputStreamWriter(out,
         // response.getCharacterEncoding()));
         csvFilePrinter = new CSVPrinter(writer, csvFormat);

         csvFilePrinter.printRecord(header);

         while (streamsMatchIter.hasNext()) {
            final Map<ChannelName, Pair<ChannelName, DataEvent>> match = streamsMatchIter.next();

            // ensure correct order
            Stream<String> rowStream = accessors.stream().sequential()
                  .map(accessorPair -> {
                     Pair<ChannelName, DataEvent> eventPair = match.get(accessorPair.getKey());
                     if (eventPair != null) {
                        return accessorPair.getValue().apply(eventPair.getValue());
                     } else {
                        return EMPTY_VALUE;
                     }
                  });

            csvFilePrinter.printRecord(new StreamIterable<String>(rowStream));
         }
      } catch (IOException e) {
         LOGGER.error("Could not write CSV.", e);
         exception.compareAndSet(null, e);
      } finally {
         try {
            if (writer != null) {
               writer.flush();
               writer.close();
            }

            if (csvFilePrinter != null) {
               csvFilePrinter.close();
            }
         } catch (IOException e) {
            LOGGER.error("Could not close CSV writer.", e);
            exception.compareAndSet(null, e);
         }
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }


   private void setupChannelColumns(DAQQueryElement daqQuery, BackendQuery backendQuery, ChannelName channelName,
         Collection<String> header, Collection<Pair<ChannelName, Function<DataEvent, String>>> accessors) {
      Set<QueryField> queryFields = daqQuery.getFields();
      List<Aggregation> aggregations =
            daqQuery.getAggregation() != null ? daqQuery.getAggregation().getAggregations() : null;
      List<Extrema> extrema = daqQuery.getAggregation() != null ? daqQuery.getAggregation().getExtrema() : null;

      QueryAnalyzer queryAnalyzer = queryAnalizerFactory.apply(backendQuery);

      for (QueryField field : queryFields) {
         if (!(QueryField.value.equals(field) && queryAnalyzer.isAggregationEnabled())) {
            StringBuilder buf = new StringBuilder(3)
                  .append(channelName.getName())
                  .append(DELIMITER_CHANNELNAME_FIELDNAME)
                  .append(field.name());

            header.add(buf.toString());
            accessors.add(Pair.of(channelName, new QueryFieldStringifyer(field.getAccessor(), EMPTY_VALUE,
                  DELIMITER_ARRAY)));
         }
      }

      if (aggregations != null && queryAnalyzer.isAggregationEnabled()) {
         for (Aggregation aggregation : aggregations) {
            StringBuilder buf = new StringBuilder(5)
                  .append(channelName.getName())
                  .append(DELIMITER_CHANNELNAME_FIELDNAME)
                  .append(QueryField.value.name())
                  .append(DELIMITER_CHANNELNAME_FIELDNAME)
                  .append(aggregation.name());

            header.add(buf.toString());
            accessors.add(Pair.of(channelName, new AggregationStringifyer(aggregation.getAccessor(), EMPTY_VALUE)));
         }
      }

      if (extrema != null && queryAnalyzer.isAggregationEnabled()) {
         for (Extrema extremum : extrema) {
            for (QueryField field : queryFields) {
               Function<DataEvent, Object> accessor = extremum.getAccessor(field);
               if (accessor != null) {
                  StringBuilder buf = new StringBuilder(7)
                        .append(channelName.getName())
                        .append(DELIMITER_CHANNELNAME_FIELDNAME)
                        .append(FIELDNAME_EXTREMA)
                        .append(DELIMITER_CHANNELNAME_FIELDNAME)
                        .append(extremum.name())
                        .append(DELIMITER_CHANNELNAME_FIELDNAME)
                        .append(field.name());

                  header.add(buf.toString());
                  accessors
                        .add(Pair.of(channelName, new QueryFieldStringifyer(accessor, EMPTY_VALUE,
                              DELIMITER_ARRAY)));
               }
            }
         }
      }
   }
}
