package ch.psi.daq.queryrest.response.json;

import java.io.OutputStream;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.ServletResponse;

import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;
import com.fasterxml.jackson.databind.ser.impl.SimpleBeanPropertyFilter;
import com.fasterxml.jackson.databind.ser.impl.SimpleFilterProvider;

import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.query.model.impl.BackendQuery;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;

/**
 * Takes a Java 8 stream and writes it to the output stream provided by the {@link ServletResponse}
 * of the current request.
 */
public class JSONResponseStreamWriter implements ResponseStreamWriter {

   private static final String DATA_RESP_FIELD = "data";

   private static final Logger LOGGER = LoggerFactory.getLogger(JSONResponseStreamWriter.class);

   @Resource
   private JsonFactory jsonFactory;

   @Resource
   private ObjectMapper mapper;

   @Override
   public void respond(List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results,
         OutputStream out) throws Exception {
      AtomicReference<Exception> exception = new AtomicReference<>();
      JsonGenerator generator = jsonFactory.createGenerator(out, JsonEncoding.UTF8);

      try {
         if (results.size() > 1) {
            generator.writeStartArray();
         }

         results
               .forEach(entryy -> {
                  DAQQueryElement daqQuery = entryy.getKey();
                  Set<String> includedFields = getFields(daqQuery);
                  ObjectWriter writer = configureWriter(includedFields);

                  try {
                     generator.writeStartArray();

                     entryy.getValue()
                           /* ensure elements are sequentially written */
                           .sequential()
                           .forEach(
                                 triple -> {
                                    try {
                                       generator.writeStartObject();
                                       generator.writeFieldName(QueryField.channel.name());
                                       writer.writeValue(generator, triple.getMiddle());
                                       generator.writeFieldName(DATA_RESP_FIELD);
                                       writer.writeValue(generator, triple.getRight());
                                       generator.writeEndObject();
                                    } catch (Exception e) {
                                       LOGGER.error("Could not write channel name of channel '{}'", triple.getMiddle(),
                                             e);
                                       exception.compareAndSet(null, e);
                                    }
                                 });

                     generator.writeEndArray();
                  } catch (Exception e) {
                     LOGGER.error("Exception while writing json for '{}'", daqQuery.getChannels(), e);
                     exception.compareAndSet(null, e);
                  }
               });
      } finally {
         if (results.size() > 1) {
            generator.writeEndArray();
         }

         generator.flush();
         generator.close();
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

   protected Set<String> getFields(DAQQueryElement query) {
      Set<QueryField> queryFields = query.getFields();
      List<Aggregation> aggregations = query.getAggregation() != null ? query.getAggregation().getAggregations() : null;
      List<Extrema> extrema = query.getAggregation() != null ? query.getAggregation().getExtrema() : null;

      Set<String> includedFields =
            new LinkedHashSet<String>(queryFields.size() + (aggregations != null ? aggregations.size() : 0)
                  + (extrema != null ? extrema.size() : 0));

      for (QueryField field : queryFields) {
         includedFields.add(field.name());
      }
      if (aggregations != null) {
         for (Aggregation aggregation : aggregations) {
            includedFields.add(aggregation.name());
         }
      }
      if (extrema != null) {
         // field of ExtremaCalculator (extrema in BinnedValueCombinedDataEvent and
         // BinnedIndexCombinedDataEvent)
         includedFields.add("extrema");
      }

      // do not write channel since it is already provided as key in mapping
      includedFields.remove(QueryField.channel.name());

      return includedFields;
   }

   /**
    * Configures the writer dynamically by including the fields which should be included in the
    * response.
    * 
    * @param includedFields set of strings which correspond to the getter method names of the
    *        classes registered as a mixed-in
    * @return the configured writer that includes the specified fields
    */
   private ObjectWriter configureWriter(Set<String> includedFields) {
      SimpleFilterProvider propertyFilter = new SimpleFilterProvider();
      propertyFilter.addFilter("namedPropertyFilter", SimpleBeanPropertyFilter.filterOutAllExcept(includedFields));
      // only write the properties not excluded in the filter
      ObjectWriter writer = mapper.writer(propertyFilter);
      return writer;
   }
}
