package ch.psi.daq.test.queryrest.controller;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVRecord;
import org.junit.After;
import org.junit.Test;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;

import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationDescriptor;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.Compression;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.request.range.RequestRangeDate;
import ch.psi.daq.domain.request.range.RequestRangePulseId;
import ch.psi.daq.domain.request.range.RequestRangeTime;
import ch.psi.daq.domain.test.TestTimeUtils;
import ch.psi.daq.queryrest.controller.QueryRestController;
import ch.psi.daq.queryrest.response.csv.CSVHTTPResponse;
import ch.psi.daq.queryrest.response.csv.CSVResponseStreamWriter;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

/**
 * Tests the {@link DaqController} implementation.
 */
public class QueryRestControllerCsvTest extends AbstractDaqRestTest {

   public static final String TEST_CHANNEL = "testChannel";
   public static final String TEST_CHANNEL_01 = TEST_CHANNEL + "1";
   public static final String TEST_CHANNEL_02 = TEST_CHANNEL + "2";
   public static final String[] TEST_CHANNEL_NAMES = new String[] {TEST_CHANNEL_01, TEST_CHANNEL_02};

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testPulseRangeQuery() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.iocSeconds);
      queryFields.add(QueryField.iocMillis);
      queryFields.add(QueryField.globalSeconds);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.shape);
      queryFields.add(QueryField.eventCount);
      queryFields.add(QueryField.value);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(queryFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals(queryFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse, record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testPulseRangeQuery_Match() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 2L, 4L, 5L));
      String channel_01 = "TestChannel_Sequence_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(1L, 3L, 5L));
      String channel_02 = "TestChannel_Sequence_" + Arrays.toString(channel_02_Seq.toArray());
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 1L, 2L, 3L, 4L, 5L));
      String channel_03 = "TestChannel_Sequence_" + Arrays.toString(channel_03_Seq.toArray());
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  -1,
                  -1),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.value);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 6;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         CSVRecord record = records.remove(0);
         assertEquals(queryFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals(queryFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               if (channelSeq.get(channel).contains(pulse)) {
                  assertEquals(channel, record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
                  assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
               } else {
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
               }

            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testPulseRangeQuery_Match_NoTimeField() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 2L, 4L, 5L));
      String channel_01 = "TestChannel_Sequence_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(1L, 3L, 5L));
      String channel_02 = "TestChannel_Sequence_" + Arrays.toString(channel_02_Seq.toArray());
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 1L, 2L, 3L, 4L, 5L));
      String channel_03 = "TestChannel_Sequence_" + Arrays.toString(channel_03_Seq.toArray());
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  -1,
                  -1),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.value);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 6;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((queryFields.size() + 1) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
            assertEquals(
                  channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + QueryField.globalMillis.name(),
                  record.get(column++));
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals((queryFields.size() + 1) * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               if (channelSeq.get(channel).contains(pulse)) {
                  assertEquals(channel, record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
                  assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               } else {
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
               }

            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testPulseRangeQueries() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      String testChannel3 = "testChannel3";
      DAQQueries request = new DAQQueries(
            new DAQQueryElement(
                  new RequestRangePulseId(
                        0,
                        1),
                  channels),
            new DAQQueryElement(
                  new RequestRangePulseId(
                        0,
                        1),
                  testChannel3));
      request.setResponse(new CSVHTTPResponse());
      channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02, testChannel3);

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.iocSeconds);
      queryFields.add(QueryField.iocMillis);
      queryFields.add(QueryField.globalSeconds);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.shape);
      queryFields.add(QueryField.eventCount);
      queryFields.add(QueryField.value);
      for (DAQQueryElement element : request) {
         element.setFields(queryFields);
      }

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERIES)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(queryFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals(queryFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse, record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testPulseRangeQueryWaveform() throws Exception {
      List<String> channels = Arrays.asList("XYWaveform");
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.iocSeconds);
      queryFields.add(QueryField.iocMillis);
      queryFields.add(QueryField.globalSeconds);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.shape);
      queryFields.add(QueryField.eventCount);
      queryFields.add(QueryField.value);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(queryFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals(queryFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("[8]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertTrue(record.get(column).startsWith("["));
               assertTrue(record.get(column++).endsWith("]"));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testTimeRangeQuery() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(0, 0),
                  TimeUtils.getTimeFromMillis(10, 0)),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.iocSeconds);
      queryFields.add(QueryField.iocMillis);
      queryFields.add(QueryField.globalSeconds);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.shape);
      queryFields.add(QueryField.eventCount);
      queryFields.add(QueryField.value);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(queryFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals(queryFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse, record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQuery() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      String startDate = TimeUtils.format(0);
      String endDate = TimeUtils.format(10);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.iocSeconds);
      queryFields.add(QueryField.iocDate);
      queryFields.add(QueryField.iocMillis);
      queryFields.add(QueryField.globalSeconds);
      queryFields.add(QueryField.globalDate);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.shape);
      queryFields.add(QueryField.eventCount);
      queryFields.add(QueryField.value);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(queryFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals(queryFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.format(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.format(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse, record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testExtremaAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            Ordering.asc,
            new AggregationDescriptor(AggregationType.extrema),
            TEST_CHANNEL_NAMES[0]);
      request.setResponse(new CSVHTTPResponse());

      String content = mapper.writeValueAsString(request);

      try {
         this.mockMvc
               .perform(MockMvcRequestBuilders
                     .post(QueryRestController.PATH_QUERY)
                     .contentType(MediaType.APPLICATION_JSON)
                     .content(content))
               .andDo(MockMvcResultHandlers.print());
         // .andExpect(MockMvcResultMatchers.model().attribute(
         // "exception",
         // Matchers.isA(IllegalArgumentException.class)));
         // should throw an exception
         assertTrue(false);
      } catch (Throwable e) {
         assertTrue(true);
      }
   }

   @Test
   public void testIndexAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            Ordering.asc,
            new AggregationDescriptor(AggregationType.index),
            TEST_CHANNEL_NAMES[0]);
      request.setResponse(new CSVHTTPResponse());

      String content = mapper.writeValueAsString(request);

      try {
         this.mockMvc
               .perform(MockMvcRequestBuilders
                     .post(QueryRestController.PATH_QUERY)
                     .contentType(MediaType.APPLICATION_JSON)
                     .content(content))
               .andDo(MockMvcResultHandlers.print());
         // .andExpect(MockMvcResultMatchers.model().attribute(
         // "exception",
         // Matchers.isA(IllegalArgumentException.class)));
         // should throw an exception
         assertTrue(false);
      } catch (Throwable e) {
         assertTrue(true);
      }
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregate() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01);
      long startTime = 0;
      long endTime = 99;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setAggregation(new AggregationDescriptor().setNrOfBins(2).setAggregations(aggregations));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.iocSeconds);
      queryFields.add(QueryField.iocMillis);
      queryFields.add(QueryField.globalSeconds);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.shape);
      queryFields.add(QueryField.eventCount);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((queryFields.size() + aggregations.size()) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
            for (Aggregation aggregation : aggregations) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + QueryField.value
                     + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + aggregation.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals((queryFields.size() + aggregations.size()) * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("5", record.get(column++));
               assertEquals("" + pulse + ".0", record.get(column++));
               assertEquals("" + (pulse + 2) + ".0", record.get(column++));
               assertEquals("" + (pulse + 4) + ".0", record.get(column++));
            }
            pulse += 5;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregateExtrema() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01);
      long startTime = 0;
      long endTime = 99;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);
      List<Extrema> extrema = new ArrayList<>();
      extrema.add(Extrema.minValue);
      extrema.add(Extrema.maxValue);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setAggregation(new AggregationDescriptor().setNrOfBins(2).setAggregations(aggregations)
            .setExtrema(extrema));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.iocSeconds);
      queryFields.add(QueryField.iocMillis);
      queryFields.add(QueryField.globalSeconds);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.shape);
      queryFields.add(QueryField.eventCount);
      queryFields.add(QueryField.value);
      request.setFields(queryFields);

      Set<QueryField> extremaFields = new LinkedHashSet<>();
      for (Extrema extremum : extrema) {
         for (QueryField queryField : queryFields) {
            if (extremum.getAccessor(queryField) != null) {
               extremaFields.add(queryField);
            }
         }
      }

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      // will not be included as it is an aggregation
      queryFields.remove(QueryField.value);
      try {
         long pulse = 0;
         int totalRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(
               (queryFields.size() + aggregations.size() + (extremaFields.size() * extrema.size())) * channels.size(),
               record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
            for (Aggregation aggregation : aggregations) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + QueryField.value
                     + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + aggregation.name(),
                     record.get(column++));
            }
            for (Extrema extremum : extrema) {
               for (QueryField queryField : extremaFields) {
                  assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME
                        + CSVResponseStreamWriter.FIELDNAME_EXTREMA
                        + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + extremum.name()
                        + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                        record.get(column++));
               }
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals((queryFields.size() + aggregations.size() + (extremaFields.size() * extrema.size()))
                  * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("5", record.get(column++));
               assertEquals("" + pulse + ".0", record.get(column++));
               assertEquals("" + (pulse + 2) + ".0", record.get(column++));
               assertEquals("" + (pulse + 4) + ".0", record.get(column++));

               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse + ".0", record.get(column++));

               assertEquals("" + (pulse + 4), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId((pulse + 4))), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId((pulse + 4))),
                     record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + (pulse + 4) + ".0", record.get(column++));
            }
            pulse += 5;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01);
      long startTime = 0;
      long endTime = 999;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setAggregation(new AggregationDescriptor().setDurationPerBin(100).setAggregations(aggregations));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.pulseId);
      queryFields.add(QueryField.iocSeconds);
      queryFields.add(QueryField.iocMillis);
      queryFields.add(QueryField.globalSeconds);
      queryFields.add(QueryField.globalMillis);
      queryFields.add(QueryField.shape);
      queryFields.add(QueryField.eventCount);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 10;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((queryFields.size() + aggregations.size()) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (QueryField queryField : queryFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + queryField.name(),
                     record.get(column++));
            }
            for (Aggregation aggregation : aggregations) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + QueryField.value
                     + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + aggregation.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalRows; ++row) {
            record = records.get(row);

            assertEquals((queryFields.size() + aggregations.size()) * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("10", record.get(column++));
               assertEquals("" + pulse + ".0", record.get(column++));
               assertEquals("" + (pulse + 4.5), record.get(column++));
               assertEquals("" + (pulse + 9) + ".0", record.get(column++));
            }
            pulse += 10;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testQuery_NoTimeFields_01() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      queryFields.add(QueryField.value);
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalRows + 1, records.size());
         // remove header
         records.remove(0);

         for (int row = 0; row < totalRows; ++row) {
            CSVRecord record = records.get(row);

            assertEquals((queryFields.size() + 1) * channels.size(), record.size());

            int column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testGzipFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setResponse(new CSVHTTPResponse(Compression.GZIP));

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.gz"));
   }

   @Test
   public void testCvsFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setResponse(new CSVHTTPResponse());

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.csv"));
   }
}
